﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Reflection.Emit;
using System.IO;

namespace caldotnetc
{
    class Emitter
    {
        private AssemblyBuilder assemblyb;
        private TypeBuilder classb;
        private string outputFile;
        private ILGenerator ilGenerator;
        private string activeFunction;

        private Dictionary<string, ILGenerator> functions = new Dictionary<string,ILGenerator>();

        public enum MathOpCode { Add, Sub, Mul, Div, And, Or, Rem };

        public Emitter(string projectName, string outputFile)
        {
            //TODO: validate projectName ?
            AssemblyName assemblyName;
            ModuleBuilder moduleb;

            this.outputFile = outputFile;
            assemblyName = new AssemblyName(projectName);
            assemblyb = AppDomain.CurrentDomain.DefineDynamicAssembly(
                assemblyName, AssemblyBuilderAccess.Save, Path.GetDirectoryName(outputFile));

            moduleb = assemblyb.DefineDynamicModule(Path.GetFileName(outputFile));
            classb = moduleb.DefineType(projectName, TypeAttributes.Public | TypeAttributes.Class | TypeAttributes.BeforeFieldInit);

            emitFunction(".main", typeof(void), null);
        }

        public void Commit()
        {
            Type t = classb.CreateType();

            MethodInfo entryp = t.GetMethod(".main");

            assemblyb.SetEntryPoint(entryp, PEFileKinds.ConsoleApplication);
            assemblyb.Save(Path.GetFileName(outputFile));
        }

        public string ActiveFunction
        {
            set
            {
                if (!functions.ContainsKey(value))
                    throw new Exception("Function " + value + " is not defined in emitter.");

                ilGenerator = functions[value];
                activeFunction = value;
            }
            get
            {
                return activeFunction;
            }
        }

        public object emitFunction(string name, Type retType, Type[] parameters)
        {
            MethodBuilder methodb;
            methodb = classb.DefineMethod(name,
                MethodAttributes.HideBySig | MethodAttributes.Static | MethodAttributes.Public,
                retType, parameters);

            functions.Add(name, methodb.GetILGenerator());

            return methodb.GetBaseDefinition();
        }

        public void emitRet()
        {
            ilGenerator.Emit(OpCodes.Ret);
        }

        public void emitWriteLine(string text)
        {
            ilGenerator.EmitWriteLine(text);
        }

        public int emitDeclareLocal(Type type)
        {
            return ilGenerator.DeclareLocal(type).LocalIndex;
        }

        public object emitDeclareGlobal(string name, Type type)
        {
            FieldBuilder fb = classb.DefineField(name, type, FieldAttributes.Static | FieldAttributes.Public);
            return fb;
        }

        public void emitNewArr(int size, Type type)
        {
            emitLDInt(size);
            ilGenerator.Emit(OpCodes.Newarr, type);
        }

        public void emitLDElem(Type type)
        {
            ilGenerator.Emit(OpCodes.Ldelem, type); 
        }

        public void emitSTElem(Type type)
        {
            ilGenerator.Emit(OpCodes.Stelem, type);
        }

        public void emitSTElemRef()
        {
            ilGenerator.Emit(OpCodes.Stelem_Ref);
        }

        public void emitLDChar(char c)
        {
            ilGenerator.Emit(OpCodes.Ldc_I4_S, (byte)c);
        }

        public void emitLDInt(int i)
        {
            switch (i)
            {
                case 0: ilGenerator.Emit(OpCodes.Ldc_I4_0); break;
                case 1: ilGenerator.Emit(OpCodes.Ldc_I4_1); break;
                case 2: ilGenerator.Emit(OpCodes.Ldc_I4_2); break;
                case 3: ilGenerator.Emit(OpCodes.Ldc_I4_3); break;
                case 4: ilGenerator.Emit(OpCodes.Ldc_I4_4); break;
                case 5: ilGenerator.Emit(OpCodes.Ldc_I4_5); break;
                case 6: ilGenerator.Emit(OpCodes.Ldc_I4_6); break;
                case 7: ilGenerator.Emit(OpCodes.Ldc_I4_7); break;
                case 8: ilGenerator.Emit(OpCodes.Ldc_I4_8); break;
                default: ilGenerator.Emit(OpCodes.Ldc_I4, (int)i); break;
            }
        }

        public void emitLDStr(string s)
        {
            ilGenerator.Emit(OpCodes.Ldstr, s);
        }

        public void emitLDFloat(float f)
        {
            ilGenerator.Emit(OpCodes.Ldc_R4, f);
        }

        public void emitLDDouble(double d)
        {
            ilGenerator.Emit(OpCodes.Ldc_R8, d);
        }

        public void emitLDLoc(int varindex)
        {
            switch (varindex)
            {
                case 0: ilGenerator.Emit(OpCodes.Ldloc_0); break;
                case 1: ilGenerator.Emit(OpCodes.Ldloc_1); break;
                case 2: ilGenerator.Emit(OpCodes.Ldloc_2); break;
                case 3: ilGenerator.Emit(OpCodes.Ldloc_3); break;
                default:
                    if (varindex > 255)
                        ilGenerator.Emit(OpCodes.Ldloc, varindex);
                    else
                        ilGenerator.Emit(OpCodes.Ldloc_S, (byte)varindex);
                    break;
            }
        }

        public void emitLDLocRef(int varindex)
        {
            if (varindex > 255)
                ilGenerator.Emit(OpCodes.Ldloca, varindex);
            else
                ilGenerator.Emit(OpCodes.Ldloca_S, (byte)varindex);
        }

        public void emitLDLocGlobal(object varhandle)
        {
            ilGenerator.Emit(OpCodes.Ldsfld, (FieldInfo)varhandle);
        }

        public void emitSTLocGlobal(object varhandle)
        {
            ilGenerator.Emit(OpCodes.Stsfld, (FieldInfo)varhandle);
        }

        public void emitSTLoc(int varindex)
        {
            switch (varindex)
            {
                case 0: ilGenerator.Emit(OpCodes.Stloc_0); break;
                case 1: ilGenerator.Emit(OpCodes.Stloc_1); break;
                case 2: ilGenerator.Emit(OpCodes.Stloc_2); break;
                case 3: ilGenerator.Emit(OpCodes.Stloc_3); break;
                default:
                    if (varindex > 255)
                        ilGenerator.Emit(OpCodes.Stloc, varindex);
                    else
                        ilGenerator.Emit(OpCodes.Stloc_S, (byte)varindex);
                    break;
            }
        }

        public void emitLDArg(int varindex)
        {
            switch (varindex)
            {
                case 0: ilGenerator.Emit(OpCodes.Ldarg_0); break;
                case 1: ilGenerator.Emit(OpCodes.Ldarg_1); break;
                case 2: ilGenerator.Emit(OpCodes.Ldarg_2); break;
                case 3: ilGenerator.Emit(OpCodes.Ldarg_3); break;
                default:
                    if (varindex > 255)
                        ilGenerator.Emit(OpCodes.Ldarg, varindex);
                    else
                        ilGenerator.Emit(OpCodes.Ldarg_S, (byte)varindex);
                    break;
            }
        }

        public void emitSTArg(int varindex)
        {
            ilGenerator.Emit(OpCodes.Starg_S, (byte)varindex);
        }

        public void emitLDInd(Type t)
        {
            if (t == typeof(int))
                ilGenerator.Emit(OpCodes.Ldind_I4);
            else
                if (t == typeof(double))
                    ilGenerator.Emit(OpCodes.Ldind_R8);
                else
                    if (t == typeof(char))
                        ilGenerator.Emit(OpCodes.Ldind_I1);
                    else
                        ilGenerator.Emit(OpCodes.Ldind_Ref);
        }

        public void emitSTInd(Type t)
        {
            if (t == typeof(int))
                ilGenerator.Emit(OpCodes.Stind_I4);
            else
                if (t == typeof(double))
                    ilGenerator.Emit(OpCodes.Stind_R8);
                else
                    if (t == typeof(char))
                        ilGenerator.Emit(OpCodes.Stind_I1);
                    else
                        ilGenerator.Emit(OpCodes.Stind_Ref);
        }

        public void emitCall(object functionHandle)
        {
            ilGenerator.EmitCall(OpCodes.Call, functionHandle as MethodInfo, null); 
        }

        public void emitCallVirt(object functionHandle)
        {
            ilGenerator.EmitCall(OpCodes.Callvirt, functionHandle as MethodInfo, null);
        }

        public void emitEQ()
        {
            ilGenerator.Emit(OpCodes.Ceq);
        }

        public void emitLogicalNot()
        {
            ilGenerator.Emit(OpCodes.Ldc_I4, (int)0);
            ilGenerator.Emit(OpCodes.Ceq);
        }

        public void emitGT()
        {
            ilGenerator.Emit(OpCodes.Cgt);
        }

        public void emitLT()
        {
            ilGenerator.Emit(OpCodes.Clt);
        }

        public object defineLabel()
        {
            return (object)ilGenerator.DefineLabel();
        }

        public void markLabel(object labelHandle)
        {
            ilGenerator.MarkLabel((Label)labelHandle);
        }

        public void emitBRtrue(object labelHandle)
        {
            ilGenerator.Emit(OpCodes.Brtrue, (Label)labelHandle);
        }

        public void emitBRfalse(object labelHandle)
        {
            ilGenerator.Emit(OpCodes.Brfalse, (Label)labelHandle);
        }

        public void emitBR(object labelHandle)
        {
            ilGenerator.Emit(OpCodes.Br, (Label)labelHandle);
        }

        public void emitConvR8()
        {
            ilGenerator.Emit(OpCodes.Conv_R8);
        }

        public void emitConvU2()
        {
            ilGenerator.Emit(OpCodes.Conv_U2);
        }

        public void emitConvI4()
        {
            ilGenerator.Emit(OpCodes.Conv_I4);
        }

        public void emidLDLoc(int varindex)
        {
            if (varindex > 255)
                ilGenerator.Emit(OpCodes.Ldloc, varindex);
            else
                ilGenerator.Emit(OpCodes.Ldloc_S, (byte)varindex);
        }

        public void emidLDLocRef(int varindex)
        {
            if (varindex > 255)
                ilGenerator.Emit(OpCodes.Ldloca, varindex);
            else
                ilGenerator.Emit(OpCodes.Ldloca_S, (byte)varindex);
        }

        public void emitPop()
        {
            ilGenerator.Emit(OpCodes.Pop);
        }

        public void emitMath(MathOpCode opc)
        {
            OpCode op = OpCodes.Nop;

            switch (opc)
            {
                case MathOpCode.Add: op = OpCodes.Add; break;
                case MathOpCode.Sub: op = OpCodes.Sub; break;
                case MathOpCode.Mul: op = OpCodes.Mul; break;
                case MathOpCode.Div: op = OpCodes.Div; break;
                case MathOpCode.And: op = OpCodes.And; break;
                case MathOpCode.Or: op = OpCodes.Or; break;
                case MathOpCode.Rem: op = OpCodes.Rem; break;
            }

            ilGenerator.Emit(op);
        }

        public void emitNewObj(ConstructorInfo ci)
        {
            ilGenerator.Emit(OpCodes.Newobj, ci);
        }

        public void emitBox(Type fromType)
        {
            ilGenerator.Emit(OpCodes.Box, fromType);
        }

        public void emitUnbox(Type toType)
        {
            ilGenerator.Emit(OpCodes.Unbox_Any, toType);
        }

        public void emitLDNull()
        {
            ilGenerator.Emit(OpCodes.Ldnull);
        }
    }
}