﻿using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;

namespace caldotnetc
{
    public class Parser
    {
        private string[] uselessTerminals = { ",", ".", ":=", "(", ")", ";", ":", 
                                              "+=", "-=", "*=", "/=", 
                                              "BEGIN", "END", "FUNCTION", "WITH", 
                                              "IF", "THEN", "REPEAT", "UNTIL",
                                              "CASE", "OF", "FOR", "USES", "WHILE", 
                                              "CONST", "DO"};

        private string[] uselessNonTerminals = { "<FunctionDeclarations>", 
                                                 "<StatementList>", 
                                                 "<VariableDeclarationList>", 
                                                 "<CaseList>", 
                                                 "<IdList>",
                                                 "<ExpressionList>",
                                                 "<ConstantDefinitionList>",
                                                 "<ArgumentList>",
                                                 "<DimensionList>",
                                                 "<Program>"};

        private SyntaxNode AST;
        private string errorString;
        private GoldParser.Grammar grammar;
        private GoldParser.Parser parser;

        public int LineNumber { get { return parser.LineNumber; } }
        public int LinePosition { get { return parser.LinePosition; } }
        public string ErrorString { get { return errorString; } }
        public string ErrorLine { get { return parser.LineText; } }
        public SyntaxNode SyntaxTree { get { return AST; } }

        public Parser(BinaryReader grammarStream)
        {
            grammar = new GoldParser.Grammar(grammarStream);
            grammarStream.Close();
        }

        public bool Parse(TextReader fileStream)
        {
            parser = new GoldParser.Parser(fileStream, grammar);
            parser.TrimReductions = true;

            try
            {
                while (true)
                {
                    switch (parser.Parse())
                    {
                        case GoldParser.ParseMessage.LexicalError:
                            errorString = string.Format("Lexical Error. Line {0}. Token {1} was not expected.", parser.LineNumber, parser.TokenText);
                            return false;

                        case GoldParser.ParseMessage.SyntaxError:
                            StringBuilder text = new StringBuilder();
                            foreach (GoldParser.Symbol tokenSymbol in parser.GetExpectedTokens())
                            {
                                text.Append(' ');
                                text.Append(tokenSymbol.ToString());
                            }
                            errorString = string.Format("Syntax Error. Line {0}. Expecting: {1}.", parser.LineNumber, text.ToString());
                            return false;

                        case GoldParser.ParseMessage.Reduction:
                            reduction();
                            break;

                        case GoldParser.ParseMessage.TokenRead:
                            tokenRead();
                            break;

                        case GoldParser.ParseMessage.Accept:
                            AST = (SyntaxNode)parser.TokenSyntaxNode;
                            trimUselessNonTerminals(AST);
                            trimEmptyNonTerminals(AST);
                            errorString = null;

                            // Root node always "<Program>"
                            if (AST.NodeType != "<Program>")
                            {
                                NonTerminalNode ntnode = new NonTerminalNode("<Program>");
                                ntnode.AppendChildNode(AST);
                                AST = ntnode;
                            }
                            return true;

                        case GoldParser.ParseMessage.InternalError:
                            errorString = "Internal Error. Something is horribly wrong.";
                            return false;

                        case GoldParser.ParseMessage.NotLoadedError:
                            errorString = "Grammar Table is not loaded.";
                            return false;

                        case GoldParser.ParseMessage.CommentError:
                            errorString = "Comment Error. Unexpected end of input.";
                            return false;

                        case GoldParser.ParseMessage.CommentBlockRead:
                        case GoldParser.ParseMessage.CommentLineRead:
                            break;
                    }
                }
            }
            finally
            {
                fileStream.Close();
            }
        }

        public bool Parse(string text)
        {
            MemoryStream ms = new MemoryStream(Encoding.ASCII.GetBytes(text));
            StreamReader sr = new StreamReader(ms);

            return Parse(sr);
        }

        private void reduction()
        {
            NonTerminalNode nonTerminal = new NonTerminalNode(parser);

            for (int i = 0; i < parser.ReductionCount; i++)
            {
                SyntaxNode node = parser.GetReductionSyntaxNode(i) as SyntaxNode;

                if (node != null)
                    nonTerminal.AppendChildNode(node);
            }

            parser.TokenSyntaxNode = nonTerminal;
        }

        private void tokenRead()
        {
            bool useless = false;

            for (int i = 0; i < uselessTerminals.Length; i++)
            {
                if (parser.TokenSymbol.Name == uselessTerminals[i])
                {
                    useless = true;
                    break;
                }
            }

            if (!useless)
            {
                TerminalNode terminal = new TerminalNode(parser);
                parser.TokenSyntaxNode = terminal;
            }
            else
                parser.TokenSyntaxNode = null;
        }

        private void trimEmptyNonTerminals(SyntaxNode tree)
        {
            NonTerminalNode ntnode = tree as NonTerminalNode;

            if ((ntnode != null) && (ntnode.Count != 0))
            {
                for (int i = 0; i < ntnode.Count; i++)
                {
                    NonTerminalNode childntnode = ntnode[i] as NonTerminalNode;
                    if (childntnode != null)
                    {
                        if (childntnode.Count != 0)
                            trimEmptyNonTerminals(childntnode);

                        if (childntnode.Count == 0)
                        {
                            ntnode.RemoveChildNode(childntnode);
                            i--;
                        }
                    }
                }
            }
        }

        private void trimUselessNonTerminals(SyntaxNode tree)
        {
            bool useless = false;

            NonTerminalNode ntnode = tree as NonTerminalNode;

            if ((ntnode == null) && (ntnode.Count == 0))
                return;

            for (int i = 0; i < ntnode.Count; i++)
            {
                NonTerminalNode childntnode = ntnode[i] as NonTerminalNode;
                if (childntnode != null)
                {
                    // trim subtree
                    trimUselessNonTerminals(childntnode);

                    useless = false;
                    for (int k = 0; k < uselessNonTerminals.Length; k++)
                    {
                        if (childntnode.NodeType == uselessNonTerminals[k])
                        {
                            useless = true;
                            break;
                        }
                    }

                    // for each useless NonTerminal
                    if (useless)
                    {
                        // cut branch
                        int index = ntnode.RemoveChildNode(childntnode);
                        i = 0;

                        // move child nodes up
                        for (int k = 0; k < childntnode.Count; k++)
                            ntnode.AppendChildNode(childntnode[k], index + k);
                    }
                }
            }
        }
    }

    abstract public class SyntaxNode
    {
        protected string nodeType;
        protected int lineNumber;
        protected int linePosition;

        public string NodeType { get { return nodeType; } }
        public override string ToString() { return NodeType; }

        public int LineNumber { get { return lineNumber; } }
        public int LinePosition { get { return linePosition; } }

        virtual public int Count { get { throw new NotSupportedException(); } }
        virtual public SyntaxNode this[int index] { get { throw new NotSupportedException(); } }
        virtual public SyntaxNode this[string type] { get { throw new NotSupportedException(); } }
        virtual public bool IsTerminal { get { return true; } }
        virtual public string Text { get { throw new NotSupportedException(); } }
    }

    public class TerminalNode : SyntaxNode
    {
        private string symbol;

        public TerminalNode(GoldParser.Parser parser)
        {
            symbol = parser.TokenText;
            nodeType = parser.TokenSymbol.ToString();
            lineNumber = parser.LineNumber;
            linePosition = parser.LinePosition;
        }

        public string Symbol { get { return symbol; } }
        override public string Text { get { return symbol; } }
    }

    public class NonTerminalNode : SyntaxNode, IEnumerable
    {
        private List<SyntaxNode> nodes = new List<SyntaxNode>();

        public NonTerminalNode(GoldParser.Parser parser)
        {
            nodeType = parser.ReductionRule.Name;
            lineNumber = parser.LineNumber;
            linePosition = parser.LinePosition;
        }

        public NonTerminalNode(string nodeType)
        {
            this.nodeType = nodeType;
        }

        override public int Count { get { return nodes.Count; } }
        override public SyntaxNode this[int index] { get { return nodes[index] as SyntaxNode; } }
        override public SyntaxNode this[string type] { get { return findFirstChildOfType(type); } }

        public void AppendChildNode(SyntaxNode node)
        {
            if (node == null)
                return;
            nodes.Add(node);
        }

        public void AppendChildNode(SyntaxNode node, int index)
        {
            if (node == null)
                return;

            nodes.Insert(index, node);
        }

        public int RemoveChildNode(SyntaxNode node)
        {
            int i;

            if (node == null)
                return -1;

            i = nodes.IndexOf(node);
            nodes.Remove(node);

            return i;
        }

        public IEnumerator GetEnumerator()
        {
            return nodes.GetEnumerator();
        }

        private SyntaxNode findFirstChildOfType(string type)
        {
            foreach (SyntaxNode n in nodes)
                if (type == n.NodeType)
                    return n;

            return null;
        }

        override public bool IsTerminal { get { return false; } }
    }
}
