﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text.RegularExpressions;

using caldotnet;

namespace caldotnetc
{
    #region THelper
    internal class THelper
    {
        public static class DummyRecFieldType { };
        public static class DummyRecType { };
        public static class DummyIgnoreType { };

        private static Emitter emitter;
        private static SymTable syms;

        public static void InitHelpers(Emitter emitter_, SymTable syms_)
        {
            emitter = emitter_;
            syms = syms_;
        }

        public static FuncSymbol findFunction(SyntaxNode node)
        {
            string fnName = node["identifier"].Text;

            FuncSymbol sym = syms[fnName] as FuncSymbol;

            if (sym == null)
            {
                sym = syms[fnName.ToLower()] as FuncSymbol;
                if ((sym == null) || !sym.InternalFunction)
                    throw new TranslatorException(node, "Function '" + fnName + "' is not defined.");
            }

            return sym;
        }

        public static int getTempVarIndex(Type vartype, SymTable syms)
        {
            string name;

            switch (vartype.Name)
            {
                case "Int32": name = "$int"; break;
                case "String": name = "$str"; break;
                case "Double": name = "$dbl"; break;
                case "Char": name = "$chr"; break;
                case "Boolean": name = "$bool"; break;

                default:
                    throw new NotImplementedException();
            }

            VarSymbol vs = syms[name] as VarSymbol;

            if (vs == null)
            {
                vs = new VarSymbol(name, emitter.emitDeclareLocal(vartype), vartype,
                    null, null, false, null, false, false, false);
                syms.Add(vs);
            }

            return vs.Index;
        }

        public static Type typeCALtoCLR(SyntaxNode nodeToShowError, string name)
        {
            switch (name.ToLower())
            {
                case "integer": return Types.Int;
                case "decimal": return Types.Double;
                case "char": return Types.Char;
                case "string": return Types.String;
                case "boolean": return Types.Bool;
                case "record": return Types.Record;
                default: throw new TranslatorException(nodeToShowError, "Unknown type: " + name);
            }
        }

        public static Type typeCALtoCLR(SyntaxNode nodeToShowError, string name, bool isArray)
        {
            if (!isArray)
                return typeCALtoCLR(nodeToShowError, name);

            switch (name.ToLower())
            {
                case "integer": return Types.IntArr;
                case "decimal": return Types.DoubleArr;
                case "char": return Types.CharArr;
                case "string": return Types.StringArr;
                case "boolean": return Types.BoolArr;
                default: throw new TranslatorException(nodeToShowError, "Unknown type: " + name);
            }
        }

        public static void ldVar(VarSymbol var)
        {
            ldVar(var, false);
        }

        public static void ldVar(VarSymbol var, bool loadByRef)
        {
            if (var.IsGlobal)
                emitter.emitLDLocGlobal(var.Handle);
            else
                if (var.IsArg)
                {
                    emitter.emitLDArg(var.Index);
                    if (var.IsByRef)
                        emitter.emitLDInd(var.Type);
                }
                else
                    if (loadByRef)
                        emitter.emitLDLocRef(var.Index);
                    else
                        emitter.emitLDLoc(var.Index);
        }

        public static void stVar(VarSymbol var)
        {
            if (var.IsGlobal)
                emitter.emitSTLocGlobal(var.Handle);
            else
                if (var.IsArg)
                {
                    if (var.IsByRef)
                        emitter.emitSTInd(var.Type);
                    else
                        emitter.emitSTArg(var.Index);
                }
                else
                    emitter.emitSTLoc(var.Index);
        }

        public static string parseStringLiteral(string str)
        {
            return new Regex(@"\\'").Replace(
                str.Substring(1, str.Length - 2),
                "'");
        }

        public static int parseIntConstant(SyntaxNode nodeToShowError, string nodeType, string nodeText)
        {
            switch (nodeType)
            {
                case "DecLiteral":
                    int i = int.Parse(nodeText);
                    return i;
                case "HexLiteral":
                    int i1 = int.Parse(nodeText.Substring(1, nodeText.Length - 1),
                        System.Globalization.NumberStyles.HexNumber);
                    return i1;
                default:
                    throw new TranslatorException(nodeToShowError, "Unexpected node type: " + nodeType + "?");
            }
        }

        public static int getArrayDimsFromParseTree(SyntaxNode node)
        {
            return node.Count - 3;
        }

        public static void doAssign(SyntaxNode nodeToShowError, int varIndex, Type fromType, Type toType)
        {
            doTranslateType(nodeToShowError, fromType, toType);
            emitter.emitSTLoc(varIndex);
        }

        public static void doAssign(SyntaxNode nodeToShowError, VarSymbol sym, Type fromType)
        {
            doTranslateType(nodeToShowError, fromType, sym.BaseType);

            if (sym.IsArray)
                emitter.emitSTElem(sym.BaseType);
            else
                THelper.stVar(sym);
        }

        public static void doTranslateType(SyntaxNode nodeToShowError, Type fromType, Type toType)
        {
            if ((toType == fromType) || (toType == typeof(DummyIgnoreType)))
                return;

            if (toType == Types.Object)
            {
                if (fromType != Types.String) //FIXME: should check fromType.IsValueType? Wont work with strings.
                    emitter.emitBox(fromType);
                return;
            }

            string convertion = fromType.Name + ">" + toType.Name;

            switch (convertion)
            {
                case "Char>Int32": break;
                case "Char>Double": emitter.emitConvR8(); break;
                case "Int32>Double": emitter.emitConvR8(); break;
                case "Int32>Char": emitter.emitConvU2(); break;
                case "Double>Int32": emitter.emitConvI4(); break;
                case "Double>Char": emitter.emitConvU2(); break;

                default:
                    throw new TranslatorException(nodeToShowError, "Cannot convert between these types.");
            }
        }

        public static Type parseUnknownExpression(SyntaxNode nodeToShowError, Type t1, Type t2, string op)
        {
            THelper.doTranslateType(nodeToShowError, t2, t1);

            if (t1 == Types.String)
                return parseStringExpression(nodeToShowError, t1, t2, op);
            else
                if ((t1 == Types.Char) ||
                    (t1 == Types.Int) ||
                    (t1 == Types.Double))
                    return parseNumericExpression(nodeToShowError, t1, t2, op);
                else
                    if (t1 == Types.Bool)
                        return parseBoolExpression(nodeToShowError, t1, t2, op);

            throw new TranslatorException(nodeToShowError, "Unsuported type in expression");
        }

        public static Type parseStringExpression(SyntaxNode nodeToShowError, Type t1, Type t2, string op)
        {
            switch (op)
            {
                case "+":
                    emitter.emitCall(t1.GetMethod("Concat", new Type[] { t1, t2 }));
                    return Types.String;
                case "=":
                    emitter.emitCall(t1.GetMethod("op_Equality", new Type[] { t1, t2 }));
                    return Types.Bool;
                case "<>":
                    emitter.emitCall(t1.GetMethod("op_Inequality", new Type[] { t1, t2 }));
                    return Types.Bool;
                default:
                    throw new TranslatorException(nodeToShowError, "Operator not supported for this type.");
            }
        }

        public static Type parseNumericExpression(SyntaxNode nodeToShowError, Type t1, Type t2, string op)
        {
            switch (op)
            {
                case "+": emitter.emitMath(Emitter.MathOpCode.Add); return t1;
                case "-": emitter.emitMath(Emitter.MathOpCode.Sub); return t1;
                case "*": emitter.emitMath(Emitter.MathOpCode.Mul); return t1;
                case "/": emitter.emitMath(Emitter.MathOpCode.Div); return t1;
                case "mod": emitter.emitMath(Emitter.MathOpCode.Rem); return t1;
                case "div":
                    emitter.emitMath(Emitter.MathOpCode.Div);
                    emitter.emitConvI4();
                    return Types.Int;
                case "=":
                    emitter.emitEQ();
                    return Types.Bool;
                case "<>":
                    emitter.emitEQ();
                    emitter.emitLogicalNot();
                    return Types.Bool;
                case "<":
                    emitter.emitLT();
                    return Types.Bool;
                case ">":
                    emitter.emitGT();
                    return Types.Bool;
                case "<=":
                    emitter.emitGT();
                    emitter.emitLogicalNot();
                    return Types.Bool;
                case ">=":
                    emitter.emitLT();
                    emitter.emitLogicalNot();
                    return Types.Bool;
                default:
                    throw new TranslatorException(nodeToShowError, "Operator not supported for this type.");
            }
        }

        public static string parseRecordName(string name)
        {
            if (name[0] == '"')
                return name.Substring(1, name.Length - 2);
            else
                return name;
        }

        public static Type parseBoolExpression(SyntaxNode nodeToShowError, Type t1, Type t2, string op)
        {
            switch (op)
            {
                case "=":
                    emitter.emitEQ();
                    return Types.Bool;
                case "<>":
                    emitter.emitEQ();
                    emitter.emitLogicalNot();
                    return Types.Bool;
                default:
                    throw new TranslatorException(nodeToShowError, "Operator not supported for this type.");
            }
        }

        public static Type parseConstant(SyntaxNode nodeToShowError, string nodeType, string nodeText)
        {
            switch (nodeType)
            {
                case "DecLiteral":
                    int i = int.Parse(nodeText);
                    emitter.emitLDInt(i);
                    return Types.Int;
                case "StringLiteral":
                    string s = THelper.parseStringLiteral(nodeText);
                    emitter.emitLDStr(s);
                    return Types.String;
                case "FloatLiteral":
                    double f = double.Parse(nodeText, System.Globalization.CultureInfo.InvariantCulture);
                    emitter.emitLDDouble(f);
                    return Types.Double;
                case "HexLiteral":
                    int i1 = int.Parse(nodeText.Substring(1, nodeText.Length - 1),
                        System.Globalization.NumberStyles.HexNumber);
                    emitter.emitLDInt(i1);
                    return Types.Int;
                case "CharLiteral":
                    char c = THelper.parseStringLiteral(nodeText)[0];
                    emitter.emitLDChar(c);
                    return Types.Char;
                case "TRUE":
                    emitter.emitLDInt(1);
                    return Types.Bool;
                case "FALSE":
                    emitter.emitLDInt(0);
                    return Types.Bool;
                default:
                    throw new TranslatorException(nodeToShowError, "Unexpected node type: " + nodeType + "?");
            }
        }

        public static void parseArraySymUse(SyntaxNode varNode, out string variable, out bool isArray)
        {
            isArray = varNode["'['"] != null;

            if (isArray && (varNode["<Variable>"]["identifier"] == null))
                throw new TranslatorException(varNode, "Arrays of arrays are not supported.");

            variable = isArray ? varNode["<Variable>"]["identifier"].Text : varNode["identifier"].Text;
        }

        public static int[] parseArrayDimsDef(SyntaxNode node)
        {
            int[] dims = new int[node.Count - 4];

            for (int i = 2; i < node.Count - 2; i++)
                dims[i - 2] = THelper.parseIntConstant(node, node[i][0].NodeType, node[i][0].Text);

            return dims;
        }

    }
    #endregion

    class TranslatorException : Exception
    {
        private SyntaxNode node;

        public int LineNumber { get { return node.LineNumber; } }
        public int LinePosition { get { return node.LinePosition; } }

        public TranslatorException(SyntaxNode node, string message)
            : base(message)
        {
            this.node = node;
        }
    }

    #region SymTable
    public enum SymbolKind { variable, function, constant };

    public abstract class BaseSymbol
    {
        protected SymbolKind kind;
        protected string name;
        public SymbolKind Kind { get { return kind; } }
        public string Name { get { return name; } }
    }

    public class VarSymbol : BaseSymbol
    {
        private int index;
        private Type type;
        private Type baseType;
        private int[] arrayDimensions;
        private bool isGlobal;
        private bool isArg;
        private bool isByRef;
        private bool isRecord;
        private object handle;
        private string recordName;

        public int Index { get { return index; } }
        public Type Type { get { return type; } }
        public Type BaseType { get { return baseType; } }
        public bool IsArray { get { return arrayDimensions != null; } }
        public int[] ArrayDimensions { get { return arrayDimensions; } }
        public int TotalArraySize { get { return getTotalArrSize(); } }
        public bool IsGlobal { get { return isGlobal; } }
        public bool IsArg { get { return isArg; } }
        public bool IsByRef { get { return isByRef; } }
        public bool IsRecord { get { return isRecord; } }
        public object Handle { get { return handle; } }
        public string RecordName { get { return recordName; } }

        public VarSymbol(string name, int index, Type type, int[] arrayDimensions, string recordName,
            bool isGlobal, object handle, bool isArg, bool isByRef, bool isRecord)
        {
            kind = SymbolKind.variable;
            this.index = index;
            this.name = name;
            this.type = type;
            this.arrayDimensions = arrayDimensions;
            this.isGlobal = isGlobal;
            this.isArg = isArg;
            this.isByRef = isByRef;
            this.isRecord = isRecord;
            this.handle = handle;
            this.recordName = recordName;

            if (IsArray)
                baseType = Types.FromArrayType(Type);
            else
                baseType = type;

            if (IsRecord)
                isByRef = true;
        }

        private int getTotalArrSize()
        {
            int total = 1;
            for (int i = 0; i < arrayDimensions.Length; i++)
            {
                total *= arrayDimensions[i];
            }

            return total;
        }
    }

    public class ConstSymbol : BaseSymbol
    {
        private string nodeType;
        private string text;

        public string NodeType { get { return nodeType; } }
        public string Text { get { return text; } }

        public ConstSymbol(string name, string nodeType, string text)
        {
            kind = SymbolKind.constant;
            this.name = name;
            this.nodeType = nodeType;
            this.text = text;
        }
    }

    public class FuncSymbol : BaseSymbol
    {
        private Type retType;
        private Type[] tparams;
        private object funcHandle;
        private bool internalFunction = false;
        private SymTable args;

        public Type RetType { get { return retType; } }
        public Type[] Params { get { return tparams; } }
        public SymTable Args { get { return args; } }
        public object Handle { get { return funcHandle; } }
        public bool InternalFunction { get { return internalFunction; } }

        public FuncSymbol(string name, Type retType, Type[] tparams, object handle, bool internalFunction, SymTable args)
        {
            kind = SymbolKind.function;
            this.name = name;
            this.retType = retType;
            this.tparams = tparams;
            this.internalFunction = internalFunction;
            this.args = args;
            funcHandle = handle;
        }
    }

    public class SymTable : ICloneable
    {
        private OrderedDictionary table = new OrderedDictionary();
        private SymTable parentTable;

        public int Count { get { return table.Count; } }

        // this does not look at parent tables! (it's only used for local ByRef params)
        public BaseSymbol this[int index] { get { return table[index] as BaseSymbol; } }

        public BaseSymbol this[string name] { get { return Find(name); } }

        public SymTable()
        {
            parentTable = null;
        }

        private SymTable(SymTable parent)
        {
            parentTable = parent;
        }

        private BaseSymbol Find(string name)
        {
            if (table.Contains(name))
                return (BaseSymbol)table[name];
            else
                return parentTable != null ? parentTable.Find(name) : null;
        }

        public SymTable MakeSubTable()
        {
            return new SymTable(this);
        }

        public bool Add(BaseSymbol sym)
        {
            if (table.Contains(sym.Name))
                return false;

            table.Add(sym.Name, sym);
            return true;
        }

        public void Clear()
        {
            table.Clear();
        }

        public Type[] ToTypeArray()
        {
            if (Count == 0)
                return null;

            Type[] types = new Type[Count];
            int i = 0;

            foreach (object key in table.Keys)
            {
                VarSymbol vs = table[key] as VarSymbol;

                if (vs == null)
                    throw new Exception("SymTable contains not only VarSymbol type elements.");

                if (vs.IsByRef)
                    types[i++] = vs.Type.MakeByRefType();
                else
                    types[i++] = vs.Type;
            }

            return types;
        }

        public bool AddFrom(SymTable tableFrom)
        {
            foreach (BaseSymbol sym in tableFrom.table.Values)
                if (!Add(sym))
                    return false;

            return true;
        }

        #region ICloneable Members

        public object Clone()
        {
            SymTable st = new SymTable();
            foreach (BaseSymbol sym in table.Values)
                st.Add(sym);

            return st;
        }

        #endregion
    }
    #endregion
}