﻿using System;
using System.Collections.Generic;

using caldotnet;

namespace caldotnetc
{
    class Translator
    {
        private delegate void NodeHandler(SyntaxNode node, SymTable syms);

        private Dictionary<string, NodeHandler> handlers;
        private SymTable syms;
        private Emitter emitter;
        private TranslatorRuntime truntime;

        private enum VarParserState { parsingGlobals, parsingArguments, parsingFunctionVars };
        private VarParserState varParserState = VarParserState.parsingGlobals;
        private SymTable varParserArgs = new SymTable();

        public Translator(TranslatorRuntime truntime, Emitter emitter)
        {
            this.truntime = truntime;
            this.emitter = emitter;
        }

        public void Translate(SyntaxNode root)
        {
            syms = new SymTable();

            syms.Add(new FuncSymbol("message", Types.Void, new Type[] { Types.String }, null, true, null));
            syms.Add(new FuncSymbol("tostr", Types.String, new Type[] { typeof(THelper.DummyIgnoreType) }, null, true, null));
            syms.Add(new FuncSymbol("strlen", Types.Int, new Type[] { Types.String }, null, true, null));
            syms.Add(new FuncSymbol("exit", Types.Void, null, null, true, null));

            THelper.InitHelpers(emitter, syms);

            handlers = new Dictionary<string, NodeHandler>();

            handlers.Add("<Program>", helperEnum);
            handlers.Add("<CompoundStatement>", helperEnum);
            handlers.Add("<FunctionDeclaration>", trFunctionDeclaration);
            handlers.Add("<ConstantDefinitions>", helperEnum);
            handlers.Add("<ConstantDef>", trConstantDef);
            handlers.Add("<VariableDeclarations>", trVariableDeclarations);
            handlers.Add("<VariableDec>", trVariableDec);
            handlers.Add("<Arg>", trVariableDec);
            handlers.Add("<UsesDeclarations>", trUsesDeclarations);
            handlers.Add("<AssignmentStatement>", trAssignmentStatement);
            handlers.Add("<AssignmentAddStatement>", trAssignmentStatement);
            handlers.Add("<AssignmentSubStatement>", trAssignmentStatement);
            handlers.Add("<AssignmentMultStatement>", trAssignmentStatement);
            handlers.Add("<AssignmentDivStatement>", trAssignmentStatement);
            handlers.Add("<FunctionCall>", trFunctionCall);
            handlers.Add("<ForStatement>", trForStatement);
            handlers.Add("<WhileStatement>", trWhileStatement);
            handlers.Add("<IfStatement>", trIfStatement);
            handlers.Add("<CaseStatement>", trCaseStatement);
            handlers.Add("<RepeatStatement>", trRepeatStatement);
            handlers.Add("<WithStatement>", trWithStatement);

            object initFunction = emitter.emitFunction("OnInit", typeof(void), null);
            emitter.ActiveFunction = "OnInit";
            truntime.LinkToRuntime();

            helperPreDeclareFunctions(root, syms);

            translate(root, syms);

            FuncSymbol entryp = syms["OnRun"] as FuncSymbol;
            if (entryp == null)
                throw new TranslatorException(root, "Entry point not defined (OnRun).");

            emitter.ActiveFunction = ".main";
            emitter.emitCall(initFunction);
            emitter.emitCall(entryp.Handle);
            emitter.emitRet();

            emitter.ActiveFunction = "OnInit";
            emitter.emitRet();

            emitter.Commit();
        }

        private void helperPreDeclareFunctions(SyntaxNode node, SymTable syms)
        {
            for (int i = 0; i < node.Count; i++) 
            {
                SyntaxNode n = node[i];

                if (n.NodeType == "<FunctionDeclaration>")
                {
                    string fnName;
                    Type retType = null;

                    n = n["<FunctionHeader>"];

                    fnName = n["identifier"].Text;
                    if (syms[fnName] != null)
                        throw new TranslatorException(n, "Symbol '" + fnName + "' already defined.");

                    varParserArgs.Clear();

                    if (n["<Arguments>"] != null)
                    {
                        varParserState = VarParserState.parsingArguments;
                        helperEnumSpec(n["<Arguments>"], syms, "<Arg>");
                        varParserState = VarParserState.parsingGlobals;
                    }

                    if (n["<TypeSpecifier>"] != null)
                    {
                        if ((n["<TypeSpecifier>"]["ARRAY"] != null) || ((n["<TypeSpecifier>"]["RECORD"] != null)))
                            throw new TranslatorException(n, "This type is not supported as return type.");

                        retType = THelper.typeCALtoCLR(n, n["<TypeSpecifier>"]["identifier"].Text);
                    }

                    if ((retType != null) && (retType != Types.String) && (retType != Types.Int) && 
                        (retType != Types.Double) && (retType != Types.Char))
                        throw new TranslatorException(n, "This type is not supported as return type.");

                    Type[] typeArray = varParserArgs.ToTypeArray();
                    object fnHandle = emitter.emitFunction(fnName, retType, typeArray);
                    syms.Add(new FuncSymbol(fnName, retType, typeArray, fnHandle, false, (SymTable)varParserArgs.Clone()));
                }
            }
        }

        private void trFunctionDeclaration(SyntaxNode node, SymTable syms)
        {
            string fnName = node["<FunctionHeader>"]["identifier"].Text;

            emitter.ActiveFunction = fnName;

            SymTable localSyms = syms.MakeSubTable();

            FuncSymbol fn = syms[fnName] as FuncSymbol;

            if (fn.Params != null)
                if (!localSyms.AddFrom(fn.Args))
                    throw new TranslatorException(node, "Symbol already defined.");

            varParserState = VarParserState.parsingFunctionVars;

            translate(node["<VariableDeclarations>"], localSyms);
            helperEnum(node["<CompoundStatement>"], localSyms);

            varParserState = VarParserState.parsingGlobals;

            emitter.emitRet();
        }

        private void trFunctionCall(SyntaxNode node, SymTable syms)
        {
            VarSymbol vsym = syms[node["identifier"].Text] as VarSymbol;

            if ((vsym != null) && (vsym.IsRecord))
                parseRecordUsage(node, syms, vsym, true, true);
            else
                doFunctionCall(node, syms, true);
        }

        private Type[] parseActuals(SyntaxNode node, SymTable syms, Type[] expectedParams, SymTable args, bool boxParams)
        {
            Type[] actuals = null;
            int arrHandle = -1;

            if (node["<Actuals>"] != null)
            {
                actuals = new Type[node["<Actuals>"].Count];

                if ((expectedParams == null) || (expectedParams.Length != node["<Actuals>"].Count))
                    throw new TranslatorException(node, "Invalid number of arguments.");


                if (boxParams)
                {
                    arrHandle = emitter.emitDeclareLocal(Types.ObjectArr);
                    emitter.emitNewArr(expectedParams.Length, Types.Object);
                    emitter.emitSTLoc(arrHandle);
                }

                for (int i = 0; i < node["<Actuals>"].Count; i++)
                {
                    if (boxParams)
                    {
                        emitter.emitLDLoc(arrHandle);
                        emitter.emitLDInt(i);
                    }

                    bool isByRef = args == null ? false : (args[i] as VarSymbol).IsByRef;
                    Type t = parseExpression(node["<Actuals>"][i], syms, isByRef);

                    if (isByRef)
                        t = t.MakeByRefType();

                    THelper.doTranslateType(node["<Actuals>"][i], t, expectedParams[i]);

                    if (boxParams)
                    {
                        THelper.doTranslateType(node["<Actuals>"][i], expectedParams[i], Types.Object);
                        emitter.emitSTElemRef();
                    }

                    actuals[i] = t;
                }
            }
            else
            {
                if (expectedParams != null)
                    throw new TranslatorException(node, "Invalid number of arguments.");
            }

            if (boxParams)
                emitter.emitLDLoc(arrHandle);

            return actuals;
        }

        private Type[] parseActualsVarArgs(SyntaxNode node, SymTable syms, Type[] expectedParams,
            int minParams, int maxParams, string recName, bool recsMustMatch)
        {
            Type[] actuals = null;
            string field;
            string rec;
            VarSymbol vsym;

            if (node["<Actuals>"] != null)
            {
                actuals = new Type[node["<Actuals>"].Count];

                if ((actuals.Length < minParams) || (actuals.Length > maxParams))
                    throw new TranslatorException(node["<Actuals>"], "Invalid number of arguments.");

                for (int i = 0; i < node["<Actuals>"].Count; i++)
                {
                    SyntaxNode lnode = node["<Actuals>"][i];

                    if (expectedParams[i] == typeof(THelper.DummyRecFieldType))
                    {
                        field = THelper.parseRecordName(lnode["identifier"].Text);
                        truntime.FieldExists(lnode, recName, field);
                        emitter.emitLDStr(field);

                    }else
                    if (expectedParams[i] == typeof(THelper.DummyRecType))
                    {
                        rec = lnode["identifier"].Text;
                        vsym = syms[rec] as VarSymbol;

                        if ((vsym == null) || (vsym.IsRecord == false))
                            throw new TranslatorException(lnode, "Invalid parameter.");

                        if (recsMustMatch)
                            if (vsym.RecordName != recName)
                                throw new TranslatorException(lnode, "Record types must match.");

                        THelper.ldVar(vsym);
                    }
                    else
                    {
                        Type t = parseExpression(lnode, syms, false);
                        THelper.doTranslateType(lnode, t, expectedParams[i]);
                        actuals[i] = t;
                    }
                }
            }
            else
            {
                if (minParams > 0)
                    throw new TranslatorException(node, "Invalid number of arguments.");
            }

            return actuals;
        }

        private void doFunctionCall(SyntaxNode node, SymTable syms, bool handleReturnStack)
        {
            FuncSymbol function = THelper.findFunction(node);
            Type[] actuals;

            if (function.Name.ToLower() != "exit")
            {
                actuals = parseActuals(node, syms, function.Params, function.Args, false);
            }
            else
            {
                function = syms[emitter.ActiveFunction] as FuncSymbol;

                if (((function.RetType == Types.Void) && (node["<Actuals>"] != null)) ||
                    (((function.RetType != Types.Void) && 
                    ((node["<Actuals>"] == null) || (node["<Actuals>"].Count != 1)))))
                    throw new TranslatorException(node, "Invalid number of arguments.");

                if (node["<Actuals>"] != null)
                {
                    Type t = parseExpression(node["<Actuals>"][0], syms);
                    THelper.doTranslateType(node["<Actuals>"][0], t, function.RetType);
                }

                emitter.emitRet();
                return;
            }

            switch (function.Name.ToLower())
            {
                case "message":
                    Type t = typeof(Console);
                    emitter.emitCall(t.GetMethod("WriteLine", new Type[] { Types.String }));
                    break;
                case "tostr":
                    if (actuals[0] == Types.String)
                        throw new TranslatorException(node, "Symbol type is already a string.");

                    int idx = THelper.getTempVarIndex(actuals[0], syms);
                    emitter.emitSTLoc(idx);
                    emitter.emitLDLocRef(idx);
                    emitter.emitCall(actuals[0].GetMethod("ToString", new Type[] { }));
                    break;
                case "strlen":
                    emitter.emitCallVirt(Types.String.GetMethod("get_Length", new Type[] { }));
                    break;
                default:
                    FuncSymbol fn = syms[function.Name] as FuncSymbol;

                    if (fn == null)
                        throw new TranslatorException(node, "Function " + function.Name + " not defined");

                    emitter.emitCall(fn.Handle);
                    break;
            }

            if (handleReturnStack)
                if ((function.RetType != null) && (function.RetType != Types.Void))
                    emitter.emitPop();
        }

        private void trConstantDef(SyntaxNode node, SymTable syms)
        {
            string name = node["identifier"].Text;
            bool hasMin = (node["'-'"] as SyntaxNode) != null;
            node = node["<constant>"][0];
            string text = node.Text;

            if (hasMin)
            {
                if ((node.NodeType != "DecLiteral") && (node.NodeType != "FloatLiteral"))
                    throw new TranslatorException(node, "Syntax error.");

                text = '-' + text;
            }

            syms.Add(new ConstSymbol(name, node.NodeType, text));
        }

        private void trVariableDeclarations(SyntaxNode node, SymTable syms)
        {
            helperEnumSpec(node, syms, "<VariableDec>");
        }

        private void trVariableDec(SyntaxNode node, SymTable syms)
        {
            SyntaxNode typeNode = node["<TypeSpecifier>"][0];
            int[] arrayDimensions = null;
            string recordName = null;
            bool isRecord = false;
            Type t;

            switch (typeNode.NodeType)
            {
                case "ARRAY":
                    t = THelper.typeCALtoCLR(node, node["<TypeSpecifier>"]["<TypeSpecifier>"][0].Text, true);
                    arrayDimensions = THelper.parseArrayDimsDef(node["<TypeSpecifier>"]);
                    break;
                case "RECORD":
                    t = Types.Record; 
                    recordName = THelper.parseRecordName(node["<TypeSpecifier>"]["identifier"].Text);
                    isRecord = true;
                    break;
                default:
                    t = THelper.typeCALtoCLR(node, node["<TypeSpecifier>"][0].Text);
                    break;
            }

            for (int i = 0; i < node.Count; i++)
                if (node[i].NodeType == "identifier")
                {
                    VarSymbol sym;

                    if (varParserState == VarParserState.parsingGlobals)
                    {
                        sym = new VarSymbol(node[i].Text, -1, t, arrayDimensions, recordName,
                            true, emitter.emitDeclareGlobal(node[i].Text, t), false, false, isRecord);

                        if(!this.syms.Add(sym))
                            throw new TranslatorException(node[i], "Symbol '" + node[i].Text + "' already defined.");
                    }
                    else
                        if (varParserState == VarParserState.parsingFunctionVars)
                        {
                            sym = new VarSymbol(node[i].Text, emitter.emitDeclareLocal(t), t,
                                arrayDimensions, recordName, false, null, false, false, isRecord);

                            if (!syms.Add(sym))
                                throw new TranslatorException(node[i], "Symbol '" + node[i].Text + "' already defined.");
                        }
                        else
                        {
                            if (varParserArgs.Count >= 255)
                                throw new TranslatorException(node[i], "What kind of sick person are you?");

                            bool isByRef = node["VAR"] != null;

                            if (isByRef && ((t != Types.String) && (t != Types.Int) && (t != Types.Double) && (t != Types.Char)))
                                throw new TranslatorException(node[i], "This type is not supported to be passed by reference.");

                            sym = new VarSymbol(node[i].Text, varParserArgs.Count, t,  
                                arrayDimensions, recordName, false, null, true, isByRef, isRecord);

                            if (!varParserArgs.Add(sym))
                                throw new TranslatorException(node[i], "Symbol '" + node[i].Text + "' already defined.");
                        }

                    if (sym.IsArray && (varParserState != VarParserState.parsingArguments))
                    {
                        if (varParserState == VarParserState.parsingFunctionVars)
                        {
                            emitter.emitNewArr(sym.TotalArraySize, sym.BaseType);
                            THelper.stVar(sym);
                        }
                        else // VarParserState.parsingGlobals
                        {
                            string fn = emitter.ActiveFunction;
                            emitter.ActiveFunction = "OnInit";

                            emitter.emitNewArr(sym.TotalArraySize, sym.BaseType);
                            THelper.stVar(sym);

                            if (fn != null)
                                emitter.ActiveFunction = fn;
                        }
                    }
                    else
                    if (sym.IsRecord && (varParserState != VarParserState.parsingArguments))
                    {
                        if (varParserState == VarParserState.parsingFunctionVars)
                        {
                            if (!truntime.CreateTableInstance(recordName))
                                throw new TranslatorException(node[i], "Cannot create Record of type " + recordName + ".");

                            THelper.stVar(sym);
                        }
                        else // VarParserState.parsingGlobals
                        {
                            string fn = emitter.ActiveFunction;
                            emitter.ActiveFunction = "OnInit";

                            if (!truntime.CreateTableInstance(recordName))
                                throw new TranslatorException(node[i], "Cannot create Record of type " + recordName + ".");

                            THelper.stVar(sym);

                            if (fn != null)
                                emitter.ActiveFunction = fn;
                        }
                    }
                }
        }

        private void trUsesDeclarations(SyntaxNode node, SymTable syms)
        {
            // TODO:
        }

        private void trWithStatement(SyntaxNode node, SymTable syms)
        {
            // TODO:
        }

        private void trAssignmentStatement(SyntaxNode node, SymTable syms)
        {
            Type resType;
            bool isArray;
            string variable;
            string field = null;

            SyntaxNode varNode = node["<Variable>"];

            THelper.parseArraySymUse(varNode, out variable, out isArray);

            VarSymbol sym = syms[variable] as VarSymbol;

            if (sym == null)
                throw new TranslatorException(varNode, "Variable " + variable + " not found.");

            if (sym.IsByRef)
            {
                emitter.emitLDArg(sym.Index);
            }
            if (sym.IsArray)
            {
                THelper.ldVar(sym);
                parseLDArray(varNode, syms, sym);
            }

            if (node.NodeType == "<AssignmentStatement>")
            {
                if (sym.IsRecord)
                {
                    field = varNode["<Variable>"]["identifier"].Text;
                    THelper.ldVar(sym);

                    switch (field)
                    {
                        case "mark": 
                        case "markedonly":
                        case "ascending": break;
                        default:
                            emitter.emitLDStr(field);
                            break;
                    }
                }

                resType = parseExpression(node[1], syms);
            }else
            {
                Type t1, t2;

                if (sym.IsRecord)
                {
                    field = varNode["<Variable>"]["identifier"].Text;

                    THelper.ldVar(sym);
                    emitter.emitLDStr(field);

                    THelper.ldVar(sym);
                    t1 = truntime.GetFieldVal(varNode, sym.RecordName, field);
                }
                else
                {
                    THelper.ldVar(sym);

                    if (sym.IsArray)
                    {
                        parseLDArray(varNode, syms, sym);
                        emitter.emitLDElem(sym.BaseType);
                    }

                    t1 = sym.BaseType;
                }

                t2 = parseExpression(node[1], syms);

                resType = null;

                switch (node.NodeType)
                {
                    case "<AssignmentAddStatement>": resType = THelper.parseUnknownExpression(node, t1, t2, "+"); break;
                    case "<AssignmentSubStatement>": resType = THelper.parseUnknownExpression(node, t1, t2, "-"); break;
                    case "<AssignmentMultStatement>": resType = THelper.parseUnknownExpression(node, t1, t2, "*"); break;
                    case "<AssignmentDivStatement>": resType = THelper.parseUnknownExpression(node, t1, t2, "/"); break;
                }
            }

            if (sym.IsRecord)
            {
                switch (field)
                {
                    case "mark": truntime.CallRecFn("SetMark"); break;
                    case "markedonly": truntime.CallRecFn("SetMarkedOnly"); break;
                    case "ascending": truntime.CallRecFn("SetAscending"); break;
                    default:
                        truntime.SetFieldVal(node, sym.RecordName, field, resType);
                        break;
                }
            }
            else
                THelper.doAssign(node, sym, resType);
        }

        private void trForStatement(SyntaxNode node, SymTable syms)
        {
            Type resType;

            string variable = node[0].Text;
            VarSymbol sym = syms[variable] as VarSymbol;

            if (sym == null)
                throw new TranslatorException(node, "Variable " + variable + " not found.");

            resType = parseExpression(node[1], syms);
            THelper.doAssign(node[1], sym, resType);

            bool downto = node[2].NodeType == "DOWNTO";

            int varNo = emitter.emitDeclareLocal(Types.Int);

            resType = parseExpression(node[3], syms);
            THelper.doAssign(node[3], varNo, resType, Types.Int);

            object startLabelH = emitter.defineLabel();
            object firstStepLH = emitter.defineLabel();
            object endLabelH = emitter.defineLabel();

            emitter.emitBR(firstStepLH);
            emitter.markLabel(startLabelH);

            THelper.ldVar(sym);
            emitter.emitLDInt(1);
            emitter.emitMath(downto ? Emitter.MathOpCode.Sub : Emitter.MathOpCode.Add);
            THelper.stVar(sym);

            emitter.markLabel(firstStepLH);

            THelper.ldVar(sym);
            emitter.emidLDLoc(varNo);

            if (downto)
                emitter.emitLT();
            else
                emitter.emitGT();

            emitter.emitBRtrue(endLabelH);

            translate(node[4], syms);
            emitter.emitBR(startLabelH);

            emitter.markLabel(endLabelH);
        }

        private void trWhileStatement(SyntaxNode node, SymTable syms)
        {
            object startLabelH = emitter.defineLabel();
            object endLabelH = emitter.defineLabel();

            emitter.markLabel(startLabelH);

            if (parseExpression(node[0], syms) != Types.Bool)
                throw new TranslatorException(node[0], "Boolean expression expected.");

            emitter.emitBRfalse(endLabelH);
            translate(node[1], syms);
            emitter.emitBR(startLabelH);
            emitter.markLabel(endLabelH);
        }

        private void trIfStatement(SyntaxNode node, SymTable syms)
        {
            if (parseExpression(node[0], syms) != Types.Bool)
                throw new TranslatorException(node[0], "Boolean expression expected.");

            object elseLabelH = emitter.defineLabel();
            object endLabelH = emitter.defineLabel();

            emitter.emitBRfalse(elseLabelH);
            translate(node[1], syms);

            if (node["ELSE"] != null)
            {
                emitter.emitBR(endLabelH);
                emitter.markLabel(elseLabelH);
                translate(node[3], syms);

            }else
                emitter.markLabel(elseLabelH);

            emitter.markLabel(endLabelH);
        }

        private void trCaseStatement(SyntaxNode node, SymTable syms)
        {
            Type t = parseExpression(node[0], syms);
            bool strCompare = false;

            if (t == Types.String)
                strCompare = true;
            else 
                if ((t != Types.Int) && (t != Types.Double) && (t != Types.Char) && (t != Types.Bool))
                    throw new TranslatorException(node[0], "A value of integral type expected.");

            int varNo = emitter.emitDeclareLocal(t);
            emitter.emitSTLoc(varNo);

            object nextLabelH = null;
            object endLabelH = emitter.defineLabel();

            for (int i = 1; i < node.Count; i++)
            {
                SyntaxNode caseNode = node[i];

                if (caseNode.NodeType != "<Case>")
                    throw new Exception("Unexpected node type.");

                if (nextLabelH != null)
                {
                    emitter.markLabel(nextLabelH);
                    nextLabelH = null;
                }

                if (caseNode[0].NodeType != "ELSE")
                {
                    object codeLabelH = emitter.defineLabel();
                    nextLabelH = emitter.defineLabel();

                    for (int k = 0; k < caseNode.Count - 1; k++)
                    {
                        bool last = k == caseNode.Count - 2;
                        THelper.doTranslateType(caseNode[k], parseExpression(caseNode[k], syms), t);
                        emitter.emitLDLoc(varNo);

                        if (strCompare)
                            emitter.emitCall(t.GetMethod("op_Equality", new Type[] { t, t }));
                        else
                            emitter.emitEQ();

                        if (last)
                            emitter.emitBRfalse(nextLabelH);
                        else
                            emitter.emitBRtrue(codeLabelH);
                    }

                    emitter.markLabel(codeLabelH);
                    translate(caseNode[caseNode.Count - 1], syms);
                    emitter.emitBR(endLabelH);
                }
                else
                {
                    if (i != node.Count - 1)
                        throw new TranslatorException(node, "ELSE branch should be the last one in case statement.");

                    translate(caseNode[caseNode.Count - 1], syms);
                }
            }

            if (nextLabelH != null)
                emitter.markLabel(nextLabelH);

            emitter.markLabel(endLabelH);
        }

        private void trRepeatStatement(SyntaxNode node, SymTable syms)
        {
            object startLabelH = emitter.defineLabel();
            object endLabelH = emitter.defineLabel();

            emitter.markLabel(startLabelH);

            // HACK: grammar problem
            for (int i = 0; i < node.Count - 1; i++)
                translate(node[i], syms);

            if (parseExpression(node[node.Count - 1], syms) != Types.Bool)
                throw new TranslatorException(node[node.Count - 1], "Boolean expression expected.");

            emitter.emitBRtrue(endLabelH);
            emitter.emitBR(startLabelH);

            emitter.markLabel(endLabelH);
        }

        private void translate(SyntaxNode node, SymTable syms)
        {
            if (node == null)
                return;

            if (!handlers.ContainsKey(node.NodeType))
                throw new Exception("No handler for node type: " + node.NodeType);

            handlers[node.NodeType](node, syms);
        }

        private void helperEnum(SyntaxNode node, SymTable syms)
        {
            for (int i = 0; i < node.Count; i++)
                translate(node[i], syms);
        }

        private void helperEnumSpec(SyntaxNode node, SymTable syms, string nodeType)
        {
            for (int i = 0; i < node.Count; i++)
                if (node[i].NodeType == nodeType)
                    translate(node[i], syms);
        }

        private Type parseExpression(SyntaxNode node, SymTable syms)
        {
            return parseExpression(node, syms, false);
        }

        private Type parseExpression(SyntaxNode node, SymTable syms, bool isByRef)
        {
            Type t;

            if (isByRef && (node.NodeType != "<Variable>"))
                throw new TranslatorException(node, "Variable expected (parameter is passed by reference).");

            switch (node.NodeType)
            {
                case "<constant>":
                    return THelper.parseConstant(node[0], node[0].NodeType, node[0].Text);
                case "<FunctionCall>":
                    VarSymbol vsym = syms[node["identifier"].Text] as VarSymbol;

                    if ((vsym != null) && (vsym.IsRecord))
                    {
                        t = parseRecordUsage(node, syms, vsym, true, false);

                        if (t == null)
                            throw new TranslatorException(node, "This function does not have return parameter.");

                        return t;
                    }
                    else
                    {
                        FuncSymbol sym = THelper.findFunction(node);

                        Type retType = sym.RetType;
                        if (retType == null)
                            throw new TranslatorException(node, "This function does not have return parameter.");

                        doFunctionCall(node, syms, false);

                        return retType;
                    }
                case "<Variable>":
                    bool isArray;
                    string varName;

                    THelper.parseArraySymUse(node, out varName, out isArray);

                    VarSymbol var = null;

                    ConstSymbol cons = syms[varName] as ConstSymbol;

                    if (cons == null)
                    {
                        var = syms[varName] as VarSymbol;

                        if (var == null)
                            throw new TranslatorException(node, "Variable " + varName + " is not defined.");

                        if (!var.IsRecord)
                            THelper.ldVar(var, isByRef);

                        t = var.Type;
                    }
                    else
                    {
                        if (isByRef)
                            throw new TranslatorException(node, "Variable expected (parameter is passed by reference).");

                        t = THelper.parseConstant(node, cons.NodeType, cons.Text);
                    }

                    if (isArray)
                    {
                        if ((var != null) && var.IsArray)
                        {
                            t = parseLDArray(node, syms, var);
                            emitter.emitLDElem(t);
                            return t;
                        }else
                        if (t == Types.String)
                        {
                            if (THelper.getArrayDimsFromParseTree(node) != 1)
                                throw new TranslatorException(node, "Too many dimensions in array expression.");

                            THelper.doTranslateType(node[2], parseExpression(node[2], syms), Types.Int);
                            emitter.emitCallVirt(Types.String.GetMethod("get_Chars", new Type[] { Types.Int }));

                            return Types.Char;
                        }

                        throw new TranslatorException(node, "Not an array type.");
                    }

                    if (var.IsRecord)
                        return parseRecordUsage(node, syms, var, false, false);

                    return t;

                case "<Expression>":
                case "<SimpleExpression>":
                case "<Term>":
                    string op = node[1].Text.ToLower();
                    Type t1;
                    Type t2;

                    if ((op == "and") || (op == "or"))
                    {
                        t1 = parseExpression(node[0], syms);
                        THelper.doTranslateType(node[0], t1, Types.Bool);

                        object l1, l2;

                        l1 = emitter.defineLabel();
                        l2 = emitter.defineLabel();

                        if (op == "and")
                            emitter.emitBRfalse(l1);
                        else
                            emitter.emitBRtrue(l1);

                        t2 = parseExpression(node[2], syms);
                        THelper.doTranslateType(node[2], t2, Types.Bool);

                        emitter.emitBR(l2);
                        emitter.markLabel(l1);

                        if (op == "and")
                            emitter.emitLDInt(0);
                        else
                            emitter.emitLDInt(1);

                        emitter.markLabel(l2);

                        return Types.Bool;
                    }
                    else
                    {
                        t1 = parseExpression(node[0], syms);
                        t2 = parseExpression(node[2], syms);

                        return THelper.parseUnknownExpression(node[0], t1, t2, op);
                    }
                case "<Factor>":
                    if (node["NOT"] != null)
                    {
                        if (parseExpression(node[1], syms) != Types.Bool)
                            throw new TranslatorException(node[1], "Operator NOT can only be applied to boolean symbol.");

                        emitter.emitLogicalNot();
                        return Types.Bool;
                    }

                    return parseExpression(node[0], syms);
                case "<NegFactor>":
                    // HACK: grammar problem
                    emitter.emitLDInt(0);
                    t = parseExpression(node[1], syms);

                    if ((t != Types.Double) && (t != Types.Int))
                        throw new TranslatorException(node[1], "Cannot negate this type.");

                    emitter.emitMath(Emitter.MathOpCode.Sub);
                    return t;
                default: 
                    throw new Exception("Unexpected node in expression.");
            }
        }

        private Type parseRecordUsage(SyntaxNode node, SymTable syms, VarSymbol vsym, bool isFnCall, bool handleReturnStack)
        {
            Type retType;
            THelper.ldVar(vsym, vsym.IsByRef);
            Type[] actuals = null;
             
            if (isFnCall)
            {
                SyntaxNode lnode = node["<FunctionCall>"];
                string fnName = lnode["identifier"].Text;

                switch (fnName.ToLower())
                {
                    case "modify": retType = truntime.CallRecFn("Modify"); break;
                    case "clearmarks": retType = truntime.CallRecFn("ClearMarks"); break;
                    case "deleteall": retType = truntime.CallRecFn("DeleteAll"); break;
                    case "init": retType = truntime.CallRecFn("Init"); break;
                    case "reset": retType = truntime.CallRecFn("Reset"); break;
                    case "delete": retType = truntime.CallRecFn("Delete"); break;
                    case "insert": retType = truntime.CallRecFn("Insert"); break;
                    case "getfilters": retType = truntime.CallRecFn("GetFilters"); break;

                    case "find":
                        parseActuals(lnode, syms, new Type[] { Types.Char }, null, false);
                        retType = truntime.CallRecFn("Find");
                        break;
                    case "next":
                        parseActuals(lnode, syms, new Type[] { Types.Int }, null, false);
                        retType = truntime.CallRecFn("Next");
                        break;
                    case "setrange":
                        actuals = parseActualsVarArgs(lnode, syms,
                            new Type[] { typeof( THelper.DummyRecFieldType), Types.Object, Types.Object }, 
                            1, 3, vsym.RecordName, false);

                        if (actuals.Length == 1)
                        {
                            emitter.emitLDStr("");
                            emitter.emitLDNull();
                        }
                        else
                        if (actuals.Length == 2)
                            emitter.emitLDNull();

                        retType = truntime.CallRecFn("SetRange");
                        break;
                    case "setfilter":
                        actuals = parseActualsVarArgs(lnode, syms,
                            new Type[] { typeof(THelper.DummyRecFieldType), Types.String },
                            2, 2, vsym.RecordName, false);
                        retType = truntime.CallRecFn("SetFilter");
                        break;
                    case "getfilter":
                        actuals = parseActualsVarArgs(lnode, syms,
                            new Type[] { typeof(THelper.DummyRecFieldType) }, 1, 1, vsym.RecordName, false);
                        retType = truntime.CallRecFn("GetFilter");
                        break;
                    case "copyfilter":
                        actuals = parseActualsVarArgs(lnode, syms,
                            new Type[] { 
                                typeof(THelper.DummyRecFieldType), 
                                typeof(THelper.DummyRecType), 
                                typeof(THelper.DummyRecFieldType) },
                            3, 3, vsym.RecordName, false);
                        retType = truntime.CallRecFn("CopyFilter");
                        break;
                    case "copyfilters":
                        actuals = parseActualsVarArgs(lnode, syms,
                            new Type[] { typeof(THelper.DummyRecType) }, 1, 1, vsym.RecordName, true);
                        retType = truntime.CallRecFn("CopyFilters");
                        break;
                    case "modifyall":
                        actuals = parseActualsVarArgs(lnode, syms,
                            new Type[] { typeof(THelper.DummyRecFieldType), Types.Object }, 
                            2, 2, vsym.RecordName, false);
                        retType = truntime.CallRecFn("ModifyAll");
                        break;
                    case "get":
                        actuals = parseActuals(lnode, syms, truntime.GetRecPrimaryKey(vsym.RecordName), null, true);
                        retType = truntime.CallRecFn("Get");
                        break;
                    case "rename":
                        actuals = parseActuals(lnode, syms, truntime.GetRecPrimaryKey(vsym.RecordName), null, true);
                        retType = truntime.CallRecFn("Rename");
                        break;
                    default:
                        throw new TranslatorException(lnode, "Identifier does not exist in type Record.");
                }
            }
            else
            {
                string varName = node["<Variable>"]["identifier"].Text;
                
                switch (varName.ToLower())
                {
                    case "tablename": retType = truntime.CallRecFn("GetTableName"); break;
                    case "count": retType = truntime.CallRecFn("GetCount"); break;
                    case "isempty": retType = truntime.CallRecFn("IsEmpty"); break;
                    case "hasfilter": retType = truntime.CallRecFn("HasFilter"); break;
                    case "mark": retType = truntime.CallRecFn("GetMark"); break;
                    case "markedonly": retType = truntime.CallRecFn("GetMarkedOnly"); break;
                    case "ascending": retType = truntime.CallRecFn("GetAscending"); break;

                    case "next":
                        emitter.emitLDInt(1);
                        retType = truntime.CallRecFn("Next");
                        break;
                        
                    default:
                        retType = truntime.GetFieldVal(node, vsym.RecordName, THelper.parseRecordName(varName));
                        break;
                }
            }

            if (handleReturnStack)
                if ((retType != null) && (retType != Types.Void))
                    emitter.emitPop();

            return retType;
        }

        private Type parseLDArray(SyntaxNode node, SymTable syms, VarSymbol var)
        {
            const int startNode = 2;
            int arrayDims = THelper.getArrayDimsFromParseTree(node);
            int k;
            int ops = 0;

            if (!var.IsArray)
                throw new TranslatorException(node, "Variable is not an array.");

            if (var.ArrayDimensions.Length != arrayDims)
                throw new TranslatorException(node, "Array dimensions do not match.");

            if (arrayDims > 1)
            {
                // simulate multidimensional array
                THelper.doTranslateType(node[startNode], parseExpression(node[startNode], syms), Types.Int);

                for (k = 1; k < arrayDims; k++)
                {
                    THelper.doTranslateType(node[startNode], parseExpression(node[k + startNode], syms), Types.Int);
                    int mul = 1;

                    for (int i = 0; i < k; i++)
                        mul *= var.ArrayDimensions[i];

                    emitter.emitLDInt(mul);
                    emitter.emitMath(Emitter.MathOpCode.Mul);
                    ops++;
                }

                for (; ops > 1; ops--)
                    emitter.emitMath(Emitter.MathOpCode.Add);

                emitter.emitMath(Emitter.MathOpCode.Add);
            }
            else
                THelper.doTranslateType(node[2], parseExpression(node[2], syms), Types.Int);

            return var.BaseType;
        }
    }
}
