﻿using System;
using System.Collections.Generic;
using System.Reflection;

using caldotnet;

namespace caldotnetc
{
    class TranslatorRuntime
    {
        private CALRuntimeParams prms;
        private DBInfo dbInfo;
        private Dictionary<string, TableInfo> tables = new Dictionary<string, TableInfo>();
        private object wrapperRef = null;
        private Emitter em;

        public TranslatorRuntime(Emitter em, CALRuntimeParams prms)
        {
            this.prms = prms;
            this.em = em;

            IDBWrapper wrapper = DBWrapper.Create(
                prms.DBWrapperModuleName, prms.DBWrapperClassName, prms.DBWrapperSettings);
            dbInfo = wrapper.GetDBInfo();

            foreach (TableInfo ti in dbInfo.Tables)
                tables.Add(ti.TableName, ti);
        }

        public void LinkToRuntime()
        {
            Type t = typeof(caldotnet.DBWrapper);

            MethodInfo mi = t.GetMethod("Create", new Type[] { Types.String, Types.String, Types.String });

            wrapperRef = em.emitDeclareGlobal(".db", typeof(caldotnet.IDBWrapper));

            em.emitLDStr(prms.DBWrapperModuleName);
            em.emitLDStr(prms.DBWrapperClassName);
            em.emitLDStr(prms.DBWrapperSettings);

            em.emitCall(mi);
            em.emitSTLocGlobal(wrapperRef);
        }

        public bool CreateTableInstance(string recordName)
        {
            if (!tables.ContainsKey(recordName))
                return false;

            Type t = typeof(caldotnet.DBWrapper);

            MethodInfo mi = t.GetMethod("GetRecordInstance",
                new Type[] { typeof(caldotnet.IDBWrapper), Types.String });

            em.emitLDLocGlobal(wrapperRef);
            em.emitLDStr(recordName);

            em.emitCall(mi);

            return true;
        }

        public caldotnet.FieldInfo FieldExists(SyntaxNode nodeToShowError, string recordName, string fieldName)
        {
            caldotnet.FieldInfo fi = tables[recordName].GetField(fieldName);

            if (fi == null)
                throw new TranslatorException(nodeToShowError, string.Format("Field '{0}' does not exist in Record '{1}'.",
                    fieldName, recordName));

            return fi;
        }

        public Type[] GetRecPrimaryKey(string recordName)
        {
            if (!tables.ContainsKey(recordName))
                throw new Exception("Record does not exist.");

            TableInfo tinfo = tables[recordName];

            Type[] types = new Type[tinfo.PrimaryKeyFields.Length];

            for (int i = 0; i < types.Length; i++)
                types[i] = tinfo.GetField(tinfo.PrimaryKeyFields[i]).CALType;

            return types;
        }

        public Type GetFieldVal(SyntaxNode nodeToShowError, string recordName, string fieldName)
        {
            caldotnet.FieldInfo fi = FieldExists(nodeToShowError, recordName, fieldName);
            MethodInfo mi = typeof(Record).GetMethod("GetFieldVal");

            em.emitLDStr(fieldName);
            em.emitCallVirt(mi);
            em.emitUnbox(fi.CALType);

            return fi.CALType;
        }

        public void SetFieldVal(SyntaxNode nodeToShowError, string recordName, string fieldName, Type valType)
        {
            caldotnet.FieldInfo fi = FieldExists(nodeToShowError, recordName, fieldName);
            MethodInfo mi = typeof(Record).GetMethod("SetFieldVal");
            THelper.doTranslateType(nodeToShowError, valType, fi.CALType);
            THelper.doTranslateType(nodeToShowError, fi.CALType, Types.Object);
            em.emitCallVirt(mi);
        }

        public Type CallRecFn(string method)
        {
            MethodInfo mi = typeof(Record).GetMethod(method);

            if (mi == null)
                throw new Exception("Method not found in Record type.");

            em.emitCallVirt(mi);

            switch (method)
            {
                case "Init": return Types.Void;
                case "ClearMarks": return Types.Void;
                case "DeleteAll": return Types.Void;
                case "Reset": return Types.Void;
                case "GetTableName": return Types.String;
                case "HasFilter": return Types.Bool;
                case "Delete": return Types.Bool;
                case "GetCount": return Types.Int;
                case "Insert": return Types.Bool;
                case "IsEmpty": return Types.Bool;
                case "Modify": return Types.Bool;
                case "Next": return Types.Int;
                case "Find": return Types.Bool;
                case "SetRange": return Types.Void;
                case "SetFilter": return Types.Void;
                case "GetFilter": return Types.String;
                case "GetFilters": return Types.String;
                case "CopyFilter": return Types.Void;
                case "CopyFilters": return Types.Void;
                case "ModifyAll": return Types.Bool;
                case "GetMark": return Types.Bool;
                case "GetMarkedOnly": return Types.Bool;
                case "GetAscending": return Types.Bool;
                case "SetMark": return Types.Void;
                case "SetMarkedOnly": return Types.Void;
                case "SetAscending": return Types.Void;
                case "Get": return Types.Bool;
                case "Rename": return Types.Bool;

                //case "GetPosition":  return Types.Void; //TODO:
                //case "SetPosition":  return Types.Void;

                default:
                    throw new NotImplementedException();
            }
        }
    }
}
