﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Text;

namespace caldotnet
{
    public class Record
    {
        #region Privates
        private TableInfo table;
        private IDBWrapper wrapper;
        private DataTable data;
        private bool ascending = true;
        private int row = -1;
        private object[] key;
        private Dictionary<object[], int> marksKeyToIndex;
        private SortedList<int, object[]> marksIndexToKey;
        private bool markedOnly;
        private Dictionary<string, Filter> fieldFilters;
        private bool needUpdate = false;

        private struct Filter
        {
            public string FilterText;
            public string SQLFilter;

            public Filter(string FilterText, string SQLFilter)
            {
                this.FilterText = FilterText;
                this.SQLFilter = SQLFilter;
            }
        }

        private class ObjectArrayComparer : IEqualityComparer<object[]>
        {
            bool IEqualityComparer<object[]>.Equals(object[] hobj1, object[] hobj2)
            {
                if (hobj1.Length != hobj2.Length)
                    return false;

                for (int i = 0; i < hobj1.Length; i++)
                    if (!hobj1[i].Equals(hobj2[i]))
                        return false;

                return true;
            }

            int IEqualityComparer<object[]>.GetHashCode(object[] hobj)
            {
                int hash = 0;

                for (int i = 0; i < hobj.Length; i++)
                    hash += hobj[i].GetHashCode();

                return hash;
            }
        }
        #endregion

        #region Internal
        internal Record(IDBWrapper wrapper, TableInfo table)
        {
            this.wrapper = wrapper;
            this.table = table;

            key = new object[table.PrimaryKeyFields.Length];

            fieldFilters = new Dictionary<string, Filter>();

            marksKeyToIndex = new Dictionary<object[], int>(new ObjectArrayComparer());
            marksIndexToKey = new SortedList<int, object[]>();

            needUpdate = true;

            updateRecord();
        }

        public DataTable GetData()
        {
            if (markedOnly)
            {
                DataTable data2 = data.Clone();

                foreach (object[] key in marksKeyToIndex.Keys)
                {
                    int r = getRowByKey(key);

                    if (r != -1)
                    {
                        DataRow row = data2.NewRow();
                        for (int i = 0; i < data.Columns.Count; i++)
                            row[i] = data.Rows[r][i];

                        data2.Rows.Add(row);
                    }
                }

                return data2;
            }
            else
                return data;
        }

        internal FieldInfo[] GetFields()
        {
            return table.Fields;
        }

        private void updateMarks()
        {
            //TOMAYBEDO: this is not an optimal data struct for this purpose
            Dictionary<object[], int> marksKeyToIndexNew = 
                new Dictionary<object[], int>(new ObjectArrayComparer());

            marksIndexToKey.Clear();

            foreach (object[] key in marksKeyToIndex.Keys)
            {
                int r = getRowByKey(key);

                if (r != -1)
                {
                    marksKeyToIndexNew.Add(key, r);
                    marksIndexToKey.Add(r, key);
                }else
                    marksKeyToIndexNew.Add(key, -1);
            }

            marksKeyToIndex = marksKeyToIndexNew;
        }

        private string getTableFilter()
        {
            StringBuilder filters = new StringBuilder();

            foreach (Filter f in fieldFilters.Values)
            {
                if (filters.Length != 0)
                    filters.Append(" AND ");

                filters.Append('(');
                filters.Append(f.SQLFilter);
                filters.Append(')');
            }

            return filters.ToString();
        }

        private void updateRecord()
        {
            if (!needUpdate)
                return;

            lock (this)
            {
                StringBuilder query = new StringBuilder();

                query.Append("SELECT * FROM [");
                query.Append(GetTableName());
                query.Append(']');

                needUpdate = false;

                string filters = getTableFilter();
                if (filters.Length != 0)
                {
                    query.Append(" WHERE ");
                    query.Append(filters);
                }

                data = wrapper.Query(query.ToString());
                updateMarks();

                setRowByKey(key);
            }
        }

        private bool keyOK(int row, object[] key)
        {
            for (int k = 0; k < key.Length; k++)
                if (!data.Rows[row][table.PrimaryKeyFields[k]].Equals(key[k]))
                    return false;

            return true;
        }

        private int getRowByKey(object[] key)
        {
            //TOMAYBEDO: this is highly not optimal :)
            for (int i = 0; i < GetCount(); i++)
                if (keyOK(i, key))
                    return i;

            return -1;
        }

        private bool setRowByKey(object[] key)
        {
            int r = getRowByKey(key);

            if (r != -1)
            {
                row = r;
                return true;
            }
            else
                return false;
        }

        private void setKeyByRow(int row)
        {
            if (data.Rows.Count <= row)
                return;

            for (int i = 0; i < key.Length; i++)
                key[i] = data.Rows[row][table.PrimaryKeyFields[i]];
        }

        private int Row
        {
            get { return row; }
            set
            {
                if ((row != -1) && (row < GetCount()))
                    data.Rows[row].RejectChanges();

                row = value;

                if (row != -1)
                    setKeyByRow(row);
            }
        }

        private string getEqConditionForKey()
        {
            StringBuilder sb = new StringBuilder();

            for (int i = 0; i < key.Length; i++)
            {
                if (i != 0)
                    sb.Append(" and ");

                sb.Append("([");
                sb.Append(table.Fields[i].Name);
                sb.Append("] = @prm");

                sb.Append(i);

                sb.Append(')');
            }

            return sb.ToString();
        }
        #endregion

        #region Searching
        public bool Find(char wich)
        {
            if (needUpdate)
                updateRecord();

            if (GetCount() == 0)
                return false;

            switch (wich)
            {
                case '+':
                    if (!markedOnly)
                    {
                        if (ascending)
                            Row = GetCount() - 1;
                        else
                            Row = 0;
                    }
                    else
                    {
                        if (ascending)
                            Row = marksIndexToKey.Keys[marksIndexToKey.Count - 1];
                        else
                            Row = marksIndexToKey.Keys[0];
                    }
                    
                    return true;
                case '-':
                    if (!markedOnly)
                    {
                        if (ascending)
                            Row = 0;
                        else
                            Row = GetCount() - 1;
                    }
                    else
                    {
                        if (ascending)
                            Row = marksIndexToKey.Keys[0];
                        else
                            Row = marksIndexToKey.Keys[marksIndexToKey.Count - 1];
                    }

                    return true;
                default:
                    throw new NotImplementedException();
            }
        }

        public int Next(int steps)
        {
            int destRow;

            if (needUpdate)
                updateRecord();

            if (!markedOnly)
            {
                destRow = ascending ? Row + steps : Row - steps;

                if ((destRow < 0) || (destRow >= GetCount()))
                    return 0;

                Row = destRow;

                return steps;
            }
            else
            {
                int currListIndex = marksKeyToIndex.ContainsKey(key) ? 
                    marksIndexToKey.IndexOfKey(marksKeyToIndex[key]) : -1;

                destRow = ascending ? currListIndex + steps : currListIndex - steps;

                if ((destRow < 0) || (destRow >= marksIndexToKey.Count))
                    return 0;

                Row = marksIndexToKey.Keys[destRow];

                return steps;
            }
        }

        public object[] GetPosition()
        {
            return key;
        }

        public void SetPosition(object[] pos)
        {
            key = pos;
            setRowByKey(key);
        }

        public bool Get(object[] primaryKey)
        {
            if (needUpdate)
                updateRecord();

            return setRowByKey(primaryKey);
        }
        #endregion

        #region Core
        public object GetFieldVal(string field)
        {
            if (Row == -1)
                return null;

            return data.Rows[Row][field];
        }

        private bool isModified(string field)
        {
            return !data.Rows[Row][field, DataRowVersion.Current].Equals(data.Rows[Row][field, DataRowVersion.Original]);
        }

        public void SetFieldVal(string field, object val)
        {
            if ((Row == -1) || (data.Rows.Count == 0))
            {
                DataRow newRow = data.NewRow();
                data.Rows.Add(newRow);
                Row = data.Rows.Count - 1;
                //Init();
            }

            data.Rows[Row][field] = val;
        }

        public bool Delete()
        {
            if (Row == -1)
                return false;

            List<DBParam> prms = new List<DBParam>(key.Length);

            for (int i = 0; i < key.Length; i++)
                prms.Add(new DBParam("@prm" + i.ToString(), key[i]));

            wrapper.QueryNoResult("DELETE FROM [" + GetTableName() + "] WHERE " + getEqConditionForKey(), prms.ToArray());

            needUpdate = true;
            updateRecord();

            if (Row >= GetCount())
                Row--;

            return true;
        }

        public void DeleteAll()
        {
            StringBuilder query = new StringBuilder();

            query.Append("DELETE FROM [");
            query.Append(GetTableName());
            query.Append(']');

            string filters = getTableFilter();
            if (filters.Length != 0)
            {
                query.Append(" WHERE ");
                query.Append(filters);
            }

            wrapper.QueryNoResult(query.ToString());
            Row = -1;

            needUpdate = true;
            updateRecord();
        }

        public bool Insert()
        {
            StringBuilder query = new StringBuilder();
            StringBuilder prmsq = new StringBuilder();
            List<DBParam> prms = new List<DBParam>();

            int i;

            query.Append("INSERT INTO [");
            query.Append(GetTableName());
            query.Append("] (");

            for (i = 0; i < table.Fields.Length; i++)
            {
                if (i != 0)
                {
                    query.Append(", ");
                    prmsq.Append(", ");
                }

                query.Append('[');
                query.Append(table.Fields[i].Name);
                query.Append(']');

                string prm = "@prm" + i.ToString();
                prmsq.Append(prm);

                prms.Add(new DBParam(prm, GetFieldVal(table.Fields[i].Name)));
            }

            query.Append(") VALUES(");
            query.Append(prmsq);
            query.Append(')');

            wrapper.QueryNoResult(query.ToString(), prms.ToArray());

            needUpdate = true;
            updateRecord();

            return true;
        }

        public bool Modify()
        {
            StringBuilder query = new StringBuilder();
            List<DBParam> prms = new List<DBParam>();

            int i, k = 0;

            bool notModified = true;

            query.Append("UPDATE [");
            query.Append(GetTableName());
            query.Append("] SET ");

            for (i = 0; i < table.Fields.Length; i++)
            {
                if (isModified(table.Fields[i].Name))
                {
                    notModified = false;

                    if (k != 0)
                        query.Append(", ");

                    query.Append('[');
                    query.Append(table.Fields[i].Name);
                    query.Append("] = ");

                    string prm = "@prm" + k.ToString();
                    query.Append(prm);

                    prms.Add(new DBParam(prm, GetFieldVal(table.Fields[i].Name)));
                    k++;
                }
            }

            if (notModified)
                return true;

            query.Append(" WHERE ");

            for (i = 0; i < table.PrimaryKeyFields.Length; i++)
            {
                if (i != 0)
                    query.Append(" AND ");

                string fieldName = table.PrimaryKeyFields[i];

                query.Append('[');
                query.Append(fieldName);
                query.Append("] = ");

                string prm = "@prm" + k.ToString();
                query.Append(prm);

                prms.Add(new DBParam(prm, GetFieldVal(fieldName)));
                k++;
            }

            //System.IO.StreamWriter sw = new System.IO.StreamWriter("c:\\log.txt", true);
            //sw.WriteLine(query.ToString());
            //sw.Close();

            wrapper.QueryNoResult(query.ToString(), prms.ToArray());

            needUpdate = true;
            updateRecord();

            return true;
        }

        public bool ModifyAll(string field, object newValue)
        {
            StringBuilder query = new StringBuilder();
            List<DBParam> prms = new List<DBParam>(1);

            query.Append("UPDATE [");
            query.Append(GetTableName());
            query.Append("] SET ");

            query.Append('[');
            query.Append(field);
            query.Append("] = @prm0");

            prms.Add(new DBParam("@prm0", newValue));

            string filters = getTableFilter();
            if (filters.Length != 0)
            {
                query.Append(" WHERE ");
                query.Append(filters);
            }

            wrapper.QueryNoResult(query.ToString(), prms.ToArray());

            needUpdate = true;
            updateRecord();

            return true;
        }

        public bool Rename(object[] values)
        {
            StringBuilder query = new StringBuilder();
            List<DBParam> prms = new List<DBParam>();

            int i, k = 0;

            query.Append("UPDATE [");
            query.Append(GetTableName());
            query.Append("] SET ");

            for (i = 0; i < table.PrimaryKeyFields.Length; i++)
            {
                if (i != 0)
                    query.Append(", ");

                string fieldName = table.PrimaryKeyFields[i];

                query.Append('[');
                query.Append(fieldName);
                query.Append("] = ");

                string prm = "@prm" + k.ToString();
                query.Append(prm);

                prms.Add(new DBParam(prm, values[i]));
                k++;
            }

            query.Append(" WHERE ");

            for (i = 0; i < table.PrimaryKeyFields.Length; i++)
            {
                if (i != 0)
                    query.Append(" AND ");

                string fieldName = table.PrimaryKeyFields[i];

                query.Append('[');
                query.Append(fieldName);
                query.Append("] = ");

                string prm = "@prm" + k.ToString();
                query.Append(prm);

                prms.Add(new DBParam(prm, GetFieldVal(fieldName)));
                k++;
            }

            wrapper.QueryNoResult(query.ToString(), prms.ToArray());

            needUpdate = true;
            updateRecord();

            return true;
        }
        #endregion

        #region Misc
        public void SetAscending(bool ascending)
        {
            this.ascending = ascending;
        }

        public bool GetAscending()
        {
            return ascending;
        }

        public int GetCount()
        {
            if (needUpdate)
                updateRecord();

            return data.Rows.Count;
        }

        public string GetTableName()
        {
            return table.TableName;
        }

        public bool IsEmpty()
        {
            return GetCount() != 0;
        }

        public void Reset()
        {
            ascending = true;
            row = -1;
            ClearMarks();
            markedOnly = false;
            fieldFilters.Clear();
            needUpdate = true;
            updateRecord();
        }

        public void Init()
        {
            foreach (FieldInfo fi in table.Fields)
                SetFieldVal(fi.Name, Types.GetDefaultValue(fi.CALType));
        }
        #endregion

        #region Marks
        public void ClearMarks()
        {
            marksIndexToKey.Clear();
            marksKeyToIndex.Clear();
        }

        public void SetMark(bool mark)
        {
            if (mark)
            {
                if (!marksKeyToIndex.ContainsKey(key))
                {
                    marksKeyToIndex.Add((object[])key.Clone(), row);
                    marksIndexToKey.Add(row, (object[])key.Clone());
                }
            }
            else
            {
                if (marksKeyToIndex.ContainsKey(key))
                {
                    marksIndexToKey.Remove(marksKeyToIndex[key]);
                    marksKeyToIndex.Remove(key);

                    if (markedOnly)
                        Row = -1;
                }
            }
        }

        public bool GetMark()
        {
            return marksKeyToIndex.ContainsKey(key);
        }

        public void SetMarkedOnly(bool monly)
        {
            markedOnly = monly;
        }

        public bool GetMarkedOnly()
        {
            return markedOnly;
        }
        #endregion

        #region Filters
        public void CopyFilter(string fromField, Record fromRecord, string toField)
        {
            SetFilter(toField, fromRecord.GetFilter(fromField));
        }

        public void CopyFilters(Record fromRecord)
        {
            foreach (FieldInfo fi in table.Fields)
                SetFilter(fi.Name, fromRecord.GetFilter(fi.Name));
        }

        public string GetFilter(string field)
        {
            if (fieldFilters.ContainsKey(field))
                return fieldFilters[field].FilterText;
            else
                return null;
        }

        public string GetFilters()
        {
            StringBuilder filters = new StringBuilder();

            foreach (string name in fieldFilters.Keys)
            {
                filters.Append(name);
                filters.Append(": ");
                filters.AppendLine(fieldFilters[name].FilterText);
            }

            return filters.ToString();
        }

        public bool HasFilter()
        {
            return fieldFilters.Count > 0;
        }

        public void SetFilter(string field, string value)
        {
            //TODO: validate filter

            if ((value == null) || (value == ""))
            {
                if (fieldFilters.ContainsKey(field))
                    fieldFilters.Remove(field);
            }
            else
            {
                fieldFilters[field] = new Filter(value,
                    ConvertFilter.Convert(value, field, table.GetField(field).CALType));
            }

            needUpdate = true;
        }

        public void SetRange(string field, object fromValue, object toValue)
        {
            if (toValue == null)
            {
                SetFilter(field, "=" + fromValue.ToString());
            }
            else
            {
                SetFilter(field, fromValue.ToString() + ".." + toValue.ToString());
            }
        }
        #endregion
    }
}